"""
Provides functions for visualizing results of anomaly detection
"""

from .boundary import boundary_image, boundary_images, framed_boundary_images
from .frame import frame_by_anomalies
from .heatmap import heatmap_image, heatmap_images
from .highlight import highlighted_image, highlighted_images
from .utils import (
    blend_image,
    composite_image,
    frame_image,
    merge_images,
    normalize_matrix,
    normalize_patch_scores,
    to_numpy,
)
